/*
 * BasicDrawModes.cpp - Version 2.1 - September 12, 2020
 *
 * Example program illustrating the basic methods available
 * in Modern OpenGL for drawing.
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// Use space to toggle what image is shown.
// Use Escape or 'X' or 'x' to exit.

// These libraries are needed to link the program (Visual Studio specific)
#pragma comment(lib,"opengl32.lib")
#pragma comment(lib,"glu32.lib")
#pragma comment(lib,"glfw3.lib")
#pragma comment(lib,"glew32s.lib")
#pragma comment(lib,"glew32.lib")

// Use the static library (so glew32.dll is not needed):
#define GLEW_STATIC
#include <GL/glew.h> 
#include <GLFW/glfw3.h>

#include "LinearR4.h"

#include "GlShaderMgr.h"
bool check_for_opengl_errors();     // Function prototype (should really go in a header file)

// Enable standard input and output via printf(), etc.
// Put this include *after* the includes for glew and GLFW!
#include <stdio.h>

// *****************************
// These variables set the dimensions of the rectanglar region we wish to view.
// They are used to help form the projection matrix.
const double Xmin = 0.0, Xmax = 9.0;
const double Ymin = 0.0, Ymax = 9.0;
const double Zmin = -1.0, Zmax = 1.0;

// ****************************
// We create one shader program: it consists of a vertex shader and a fragment shader
unsigned int shaderProgram1;
const unsigned int vertPos_loc = 0;   // Corresponds to "location = 0" in the verter shader definition
const unsigned int vertColor_loc = 1; // Corresponds to "location = 1" in the verter shader definition
const char* projMatName = "projectionMatrix";		// Name of the uniform variable projectionMatrix
int projMatLocation;				                // Location of the projectionMatrix in the shader program.
const char* modelviewMatName = "modelviewMatrix";	// Name of the uniform variable modelviewMatrix
int modelviewMatLocation;			                // Location of the modelviewMatrix in the "shader program.

//  The Projection matrix: Controls the "camera view/field-of-view" transformation
// A ModelView matrix controls the placement of a particular object in 3-space.
LinearMapR4 theProjectionMatrix;		//  The Projection matrix: Controls the "camera/view" transformation
LinearMapR4 theModelViewMatrix;
float matEntries[16];		// Holds 16 floats (since cannot load doubles into a shader that uses floats)


// ************************
// General data helping with setting up VAO (Vertex Array Objects)
//    and Vertex Buffer Objects and Element Buffer Objects.
// ************************

const int NumVAOs = 4;
const int NumVBOs = 2;
const int NumEBOs = 2;
const int vaoDrawArrays = 0;
const int vaoDrawElements = 1;
const int vaoMultiDrawElements = 2;
const int vaoDrawElementsPrimitiveRestart = 3;
const int vboDrawArrays = 0;
const int vboDrawEltsAll = 1;
const int eboDrawMultiDrawElts = 0;
const int eboDrawEltsPrimRestart = 1;

unsigned int myVBO[NumVBOs];  // Vertex Buffer Object - holds an array of data
unsigned int myVAO[NumVAOs];  // Vertex Array Buffer Object - holds info about an array of vertex data;
unsigned int myEBO[NumEBOs];  // Element Array Buffer Object - holds info about an array of vertex data;

unsigned int restartPrimitiveCode = 0xffffffff;     // Max. unsigned integer - signals of a  drawing primitive (triangle strip)

// Each of the triangle strips has 8 vertices
int multidrawCount[3] = { 8, 8, 8 }; 
// The starting positions in the EBO of the three triangle strips
void* multidrawStartIdx[3] = { 0, (void*)(8 * sizeof(unsigned int)), (void*)(16 * sizeof(unsigned int)) };

// *************************
// mySetupGeometries defines the scene data, especially vertex positions and colors.
//    - It also loads all the data into the VAO's (Vertex Array Objects) and
//      into the VBO's (Vertex Buffer Objects) and
//      into the EBO's (Element Buffer Objects)
// This routine is only called once to initialize the data.
// *************************
void mySetupGeometries() {

	// We model all four images in the range [-1,-1] x [-1,1],
    //    and use the modelview matrix to place them appropriately.

    // Allocate Vertex Array Objects (VAOs) and Vertex Buffer Objects (VBOs).
    glGenVertexArrays(NumVAOs, &myVAO[0]);
    glGenBuffers(NumVBOs, &myVBO[0]);
    glGenBuffers(NumEBOs, &myEBO[0]);

    // Four identical geometries, described explicitly in arrays.
    // Each is drawn as three triangle strips with six triangles per strip.
    //  #1 - drawn with glDrawArrays.
    //  #2 - drawn with glDrawElements.
    //  #3 - drawn with glMultiDrawElements.
    //  #4 - drawn with glDrawElements with Primitive Restart index enabled.
 

    // FIRST GEOMETRY: Drawn with glDrawArrays.
    // Three triangle strips with eight vertices each.
    float firstGeom[][3] = {
        { 0.0f, 3.0f, 0.0 }, { 0.0f, 2.0f, 0.0 }, { 1.0f, 3.0f, 0.0 }, { 1.0f, 2.0f, 0.0 },
            { 2.0f, 3.0f, 0.0 }, { 2.0f, 2.0f, 0.0 }, { 3.0f, 3.0f, 0.0 }, { 3.0f, 2.0f, 0.0 },
        { 0.0f, 2.0f, 0.0 }, { 0.0f, 1.0f, 0.0 }, { 1.0f, 2.0f, 0.0 }, { 1.0f, 1.0f, 0.0 },
            { 2.0f, 2.0f, 0.0 }, { 2.0f, 1.0f, 0.0 }, { 3.0f, 2.0f, 0.0 }, { 3.0f, 1.0f, 0.0 },
        { 0.0f, 1.0f, 0.0 }, { 0.0f, 0.0f, 0.0 }, { 1.0f, 1.0f, 0.0 }, { 1.0f, 0.0f, 0.0 },
            { 2.0f, 1.0f, 0.0 }, { 2.0f, 0.0f, 0.0 }, { 3.0f, 1.0f, 0.0 }, { 3.0f, 0.0f, 0.0 },
    };

    // Bind (and initialize) the Vertex Array Object and its associated buffer objects
    glBindVertexArray(myVAO[vaoDrawArrays]);
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[vboDrawArrays]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(firstGeom), firstGeom, GL_STATIC_DRAW);
    glVertexAttribPointer(vertPos_loc, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);	// Info about where positions are in the VBO
    glEnableVertexAttribArray(vertPos_loc);									// Enable the stored vertex positions

     // OTHER THREE GEOMETRIES - vertex data is same for all three. No need to repeat vertices.
    float otherGeom[][3] = {
        { 0.0f, 3.0f, 0.0 }, { 1.0f, 3.0f, 0.0 }, { 2.0f, 3.0f, 0.0 }, { 3.0f, 3.0f, 0.0 },
        { 0.0f, 2.0f, 0.0 }, { 1.0f, 2.0f, 0.0 }, { 2.0f, 2.0f, 0.0 }, { 3.0f, 2.0f, 0.0 },
        { 0.0f, 1.0f, 0.0 }, { 1.0f, 1.0f, 0.0 }, { 2.0f, 1.0f, 0.0 }, { 3.0f, 1.0f, 0.0 },
        { 0.0f, 0.0f, 0.0 }, { 1.0f, 0.0f, 0.0 }, { 2.0f, 0.0f, 0.0 }, { 3.0f, 0.0f, 0.0 },
    };

    // EBO data for glDrawElements and glMultiDrawElements
    // Three triangle strips with eight vertices each.
    unsigned int drawmultidrawElts[] = {
        0, 4, 1, 5, 2, 6, 3, 7,
        4, 8, 5, 9, 6, 10, 7, 11,
        8, 12, 9, 13, 10, 14, 11, 15
    };

    // EBO data for glDrawElementsPrimitiveRestart
    // Three triangle strips with eight vertices each.
    // Breaks between triangle strips indicated by restartPrimitiveCode.
    unsigned int drawPrimRestartElts[] = {
        0, 4, 1, 5, 2, 6, 3, 7, restartPrimitiveCode,
        4, 8, 5, 9, 6, 10, 7, 11, restartPrimitiveCode,
        8, 12, 9, 13, 10, 14, 11, 15
    };

    // Drawing command data for glMultiDrawElements is contained in the 
    //     globally defined arrays multiDrawCount[] and multiDrawStartIdx[]
    // For these see above!

    glBindVertexArray(myVAO[vaoDrawElements]);
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[vaoDrawElements]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(otherGeom), otherGeom, GL_STATIC_DRAW);
    glVertexAttribPointer(vertPos_loc, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glEnableVertexAttribArray(vertPos_loc);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, myEBO[eboDrawMultiDrawElts]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(drawmultidrawElts), drawmultidrawElts, GL_STATIC_DRAW);

    // Do not need to give the glBufferData commands again. Data was already loaded into the buffer.
    glBindVertexArray(myVAO[vaoMultiDrawElements]);
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[vboDrawEltsAll]);
    glVertexAttribPointer(vertPos_loc, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glEnableVertexAttribArray(vertPos_loc);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, myEBO[eboDrawMultiDrawElts]);

    glBindVertexArray(myVAO[vaoDrawElementsPrimitiveRestart]);
    glBindBuffer(GL_ARRAY_BUFFER, myVBO[vboDrawEltsAll]);
    glVertexAttribPointer(vertPos_loc, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glEnableVertexAttribArray(vertPos_loc);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, myEBO[eboDrawEltsPrimRestart]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(drawPrimRestartElts), drawPrimRestartElts, GL_STATIC_DRAW);

	// This is optional, but allowed.  The VAO already knows which buffer (VBO) is holding its
	//     vertex data, so it is OK to unbind the VBO here.
	glBindBuffer(GL_ARRAY_BUFFER, 0);

	// Also optional, but perhaps a good idea so that the last VAO is not used by accident.
	glBindVertexArray(0);

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}
// *************************************
// Main routine for rendering the scene
// myRenderScene() is called every time the scene needs to be redrawn.
// mySetupGeometries() has already created the vertex and buffer objects, etc.
// my_setup_SceneData() has already created the shader programs.
// *************************************
void myRenderScene() {

	// Clear the rendering window
	static const float black[] = { 0.0f, 0.0f, 0.0f, 0.0f };
	const float clearDepth = 1.0f;
	glClearBufferfv(GL_COLOR, 0, black);
	glClearBufferfv(GL_DEPTH, 0, &clearDepth);	// Must pass in a pointer to the depth value!
    
	// This very very simple shader program is used for all the items.
	glUseProgram(shaderProgram1);

    // Draw bottom left grid with three calls to glDrawArrays
    glBindVertexArray(myVAO[vaoDrawArrays]);
    glVertexAttrib3f(vertColor_loc, 1.0f, 0.4f, 0.4f);		// A light red color
    theModelViewMatrix.Set_glTranslate(1.0, 1.0, 0.0);
    theModelViewMatrix.DumpByColumns(matEntries);
    glUniformMatrix4fv(modelviewMatLocation, 1, false, matEntries);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 8);
    glDrawArrays(GL_TRIANGLE_STRIP, 8, 8);
    glDrawArrays(GL_TRIANGLE_STRIP, 16, 8);

    // Draw bottom right grid with three calls to glDrawElements
    glBindVertexArray(myVAO[vaoDrawElements]);
    glVertexAttrib3f(vertColor_loc, 1.0f, 1.0f, 0.4f);		// A light yellow color
    theModelViewMatrix.Set_glTranslate(5.0, 1.0, 0.0);
    theModelViewMatrix.DumpByColumns(matEntries);
    glUniformMatrix4fv(modelviewMatLocation, 1, false, matEntries);
    glDrawElements(GL_TRIANGLE_STRIP, 8, GL_UNSIGNED_INT, 0);
    glDrawElements(GL_TRIANGLE_STRIP, 8, GL_UNSIGNED_INT, (void*)(8 * sizeof(unsigned int)));
    glDrawElements(GL_TRIANGLE_STRIP, 8, GL_UNSIGNED_INT, (void*)(16*sizeof(unsigned int)));

    // Draw upper left grid with a single call to glMultiDrawElements
    glBindVertexArray(myVAO[vaoMultiDrawElements]);
    glVertexAttrib3f(vertColor_loc, 0.6f, 0.6f, 1.0f);		// A light blue color
    theModelViewMatrix.Set_glTranslate(1.0, 5.0, 0.0);
    theModelViewMatrix.DumpByColumns(matEntries);
    glUniformMatrix4fv(modelviewMatLocation, 1, false, matEntries);
    glMultiDrawElements(GL_TRIANGLE_STRIP, multidrawCount, GL_UNSIGNED_INT, multidrawStartIdx, 3);

    // Draw upper right grid with a single call to glDrawElements with Primitive Restart
    glBindVertexArray(myVAO[vaoDrawElementsPrimitiveRestart]);
    glVertexAttrib3f(vertColor_loc, 0.4f, 1.0f, 0.4f);		// A light green color
    theModelViewMatrix.Set_glTranslate(5.0, 5.0, 0.0);
    theModelViewMatrix.DumpByColumns(matEntries);
    glUniformMatrix4fv(modelviewMatLocation, 1, false, matEntries);
    glEnable(GL_PRIMITIVE_RESTART);
    glPrimitiveRestartIndex(restartPrimitiveCode);
    glDrawElements(GL_TRIANGLE_STRIP, 26, GL_UNSIGNED_INT, 0);
    glDisable(GL_PRIMITIVE_RESTART);     // Not clear that there is any reason to disable (maybe for performance?)

    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void my_setup_SceneData() {
	mySetupGeometries();

    // Load the (only) shader program.
    GlShaderMgr::LoadShaderSource("BDMshaders.glsl");
    shaderProgram1 = GlShaderMgr::CompileAndLinkAll();
    
    // Get the locations of the projection and model view matrices in the shader program.
    projMatLocation = glGetUniformLocation(shaderProgram1, projMatName);
    modelviewMatLocation = glGetUniformLocation(shaderProgram1, modelviewMatName);
 
    check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

// *******************************************************
// Process all key press events.
// This routine is called each time a key is pressed or released.
// *******************************************************
void key_callback(GLFWwindow* window, int key, int scancode, int action, int mods) {
	if (action == GLFW_RELEASE) {
		return;			// Ignore key up (key release) events
	}
	if (key == GLFW_KEY_ESCAPE || key == GLFW_KEY_X) {
		glfwSetWindowShouldClose(window, true);
	}
}


// *************************************************
// This function is called with the graphics window is first created,
//    and again whenever it is resized.
// The Projection View Matrix is set here.
// *************************************************
void window_size_callback(GLFWwindow* window, int width, int height) {
	glViewport(0, 0, width, height);		// Draw into entire window

	// Setup the projection matrix as an orthographic view.
	// Determine the min and max values for x and y that should appear in the window.
	// The complication is that the aspect ratio of the window may not match the
	//		aspect ratio of the scene we want to view.
	double w = (width == 0) ? 1.0 : (double)width;
	double h = (height == 0) ? 1.0 : (double)height;

    // Center point and one-half of width and height
    double centerX = (Xmax + Xmin) / 2.0;
    double centerY = (Ymax + Ymin) / 2.0;
    double halfWidth = (Xmax - Xmin) / 2.0;
    double halfHeight = (Ymax - Ymin) / 2.0;

    // Scale (increase) either width or height up to match window aspect ration
    double scale = (halfHeight * w ) / (halfWidth * h);
    if (scale > 1.0) {
        halfWidth *= scale;
    }
    else {
        halfHeight /= scale;
    }
    double windowXmin = centerX - halfWidth;
    double windowXmax = centerX + halfWidth;
    double windowYmin = centerX - halfHeight;
    double windowYmax = centerY + halfHeight;
 
	// Using the max & min values for x & y & z that should be visible in the window,
	//		we set up the orthographic projection.
	theProjectionMatrix.Set_glOrtho(windowXmin, windowXmax, windowYmin, windowYmax, Zmin, Zmax);

	theProjectionMatrix.DumpByColumns(matEntries);
	if (glIsProgram(shaderProgram1)) {
		glUseProgram(shaderProgram1);
		glUniformMatrix4fv(projMatLocation, 1, false, matEntries);
	}
	check_for_opengl_errors();   // Really a great idea to check for errors -- esp. good for debugging!
}

void my_setup_OpenGL() {
	
	glEnable(GL_DEPTH_TEST);	// Enable depth buffering
	glDepthFunc(GL_LEQUAL);		// Useful for multipass shaders

	// Set polygon drawing mode: render triangles as lines only
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	// Disable backface culling to render both sides of triangles
	// glDisable(GL_CULL_FACE);

	// The following commands should induce OpenGL to create round points and 
	//	antialias points and lines.  (This is implementation dependent unfortunately.)
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);	// Make round points, not square points
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);		// Antialias the lines
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Specify the diameter of points, and the width of lines. (Implementation dependent.)
	//   Permissible values are 1 and greater.
	glPointSize(2);
	glLineWidth(2);
}

void error_callback(int error, const char* description)
{
	// Print error
	fputs(description, stderr);
}

void setup_callbacks(GLFWwindow* window) {
	// Set callback function for resizing the window
	glfwSetFramebufferSizeCallback(window, window_size_callback);

	// Set callback for key up/down/repeat events
	glfwSetKeyCallback(window, key_callback);

	// Set callbacks for mouse movement (cursor position) and mouse botton up/down events.
	// glfwSetCursorPosCallback(window, cursor_pos_callback);
	// glfwSetMouseButtonCallback(window, mouse_button_callback);
}

// **********************
// Here is the main program
// **********************

int main() {
	glfwSetErrorCallback(error_callback);	// Supposed to be called in event of errors. (doesn't work?)
	glfwInit();
#if defined(__APPLE__) || defined(__linux__)
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
#endif

    const int initWidth = 800;
	const int initHeight = 600;
	GLFWwindow* window = glfwCreateWindow(initWidth, initHeight, "BasicDrawModes", NULL, NULL);
	if (window == NULL) {
		fprintf(stdout, "Failed to create GLFW window!\n");
		glfwTerminate();
		return -1;
	}
	glfwMakeContextCurrent(window);

	if (GLEW_OK != glewInit()) {
		printf("Failed to initialize GLEW!.\n");
		return -1;
	}


	// Print info of GPU and supported OpenGL version
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("OpenGL version supported %s\n", glGetString(GL_VERSION));
#ifdef GL_SHADING_LANGUAGE_VERSION
	printf("Supported GLSL version is %s.\n", (char *)glGetString(GL_SHADING_LANGUAGE_VERSION));
#endif
    printf("Using GLEW version %s.\n", glewGetString(GLEW_VERSION));

	printf("------------------------------\n");
	printf("Press ESCAPE or 'X' or 'x' to exit.\n");
	
    setup_callbacks(window);

	// Initialize OpenGL, the scene and the shaders
    my_setup_OpenGL();
	my_setup_SceneData();
    window_size_callback(window, initWidth, initHeight);
 
    // Loop while program is not terminated.
	while (!glfwWindowShouldClose(window)) {
	
		myRenderScene();				// Render into the current buffer
		glfwSwapBuffers(window);		// Displays what was just rendered (using double buffering).

		// Poll events (key presses, mouse events)
		glfwWaitEvents();					// Use this if no animation.
		// glfwWaitEventsTimeout(1.0/60.0);	// Use this to animate at 60 frames/sec (timing is NOT reliable)
		// glfwPollEvents();				// Use this version when animating as fast as possible
	}

	glfwTerminate();
	return 0;
}

// If an error is found, it could have been caused by any command since the
//   previous call to check_for_opengl_errors()
// To find what generated the error, you can try adding more calls to
//   check_for_opengl_errors().
char errNames[8][36] = {
    "Unknown OpenGL error",
    "GL_INVALID_ENUM", "GL_INVALID_VALUE", "GL_INVALID_OPERATION",
    "GL_INVALID_FRAMEBUFFER_OPERATION", "GL_OUT_OF_MEMORY",
    "GL_STACK_UNDERFLOW", "GL_STACK_OVERFLOW" };
bool check_for_opengl_errors() {
    int numErrors = 0;
    GLenum err;
    while ((err = glGetError()) != GL_NO_ERROR) {
        numErrors++;
        int errNum = 0;
        switch (err) {
        case GL_INVALID_ENUM:
            errNum = 1;
            break;
        case GL_INVALID_VALUE:
            errNum = 2;
            break;
        case GL_INVALID_OPERATION:
            errNum = 3;
            break;
        case GL_INVALID_FRAMEBUFFER_OPERATION:
            errNum = 4;
            break;
        case GL_OUT_OF_MEMORY:
            errNum = 5;
            break;
        case GL_STACK_UNDERFLOW:
            errNum = 6;
            break;
        case GL_STACK_OVERFLOW:
            errNum = 7;
            break;
        }
        printf("OpenGL ERROR: %s.\n", errNames[errNum]);
    }
    return (numErrors != 0);
}